package com.squareup.anvil.compiler.codegen.dagger

import com.squareup.anvil.compiler.assistedFqName
import com.squareup.anvil.compiler.daggerDoubleCheckFqNameString
import com.squareup.anvil.compiler.daggerLazyFqName
import com.squareup.anvil.compiler.injectFqName
import com.squareup.anvil.compiler.internal.capitalize
import com.squareup.anvil.compiler.internal.reference.ClassReference
import com.squareup.anvil.compiler.internal.reference.ParameterReference
import com.squareup.anvil.compiler.internal.reference.PropertyReference
import com.squareup.anvil.compiler.internal.reference.TypeParameterReference
import com.squareup.anvil.compiler.internal.reference.Visibility.PRIVATE
import com.squareup.anvil.compiler.internal.reference.allSuperTypeClassReferences
import com.squareup.anvil.compiler.internal.reference.argumentAt
import com.squareup.anvil.compiler.internal.reference.generateClassName
import com.squareup.anvil.compiler.internal.withJvmSuppressWildcardsIfNeeded
import com.squareup.anvil.compiler.providerFqName
import com.squareup.kotlinpoet.ClassName
import com.squareup.kotlinpoet.FunSpec
import com.squareup.kotlinpoet.ParameterizedTypeName
import com.squareup.kotlinpoet.ParameterizedTypeName.Companion.parameterizedBy
import com.squareup.kotlinpoet.TypeName
import com.squareup.kotlinpoet.asClassName
import dagger.Lazy
import dagger.internal.ProviderOfLazy
import javax.inject.Provider

internal fun TypeName.wrapInProvider(): ParameterizedTypeName {
  return Provider::class.asClassName().parameterizedBy(this)
}

internal fun TypeName.wrapInLazy(): ParameterizedTypeName {
  return Lazy::class.asClassName().parameterizedBy(this)
}

internal fun List<ParameterReference>.mapToConstructorParameters(): List<ConstructorParameter> {
  return fold(listOf()) { acc, callableReference ->
    acc + callableReference.toConstructorParameter(callableReference.name.uniqueParameterName(acc))
  }
}

private fun ParameterReference.toConstructorParameter(
  uniqueName: String
): ConstructorParameter {
  val type = type()

  val isWrappedInProvider = type.asClassReferenceOrNull()?.fqName == providerFqName
  val isWrappedInLazy = type.asClassReferenceOrNull()?.fqName == daggerLazyFqName
  val isLazyWrappedInProvider = isWrappedInProvider &&
    type.unwrappedFirstType.asClassReferenceOrNull()?.fqName == daggerLazyFqName

  val typeName = when {
    isLazyWrappedInProvider -> type.unwrappedFirstType.unwrappedFirstType
    isWrappedInProvider || isWrappedInLazy -> type.unwrappedFirstType
    else -> type
  }.asTypeName().withJvmSuppressWildcardsIfNeeded(this, type)

  val assistedAnnotation = annotations.singleOrNull { it.fqName == assistedFqName }

  val assistedIdentifier = assistedAnnotation
    ?.argumentAt("value", 0)
    ?.value()
    ?: ""

  return ConstructorParameter(
    name = uniqueName,
    originalName = name,
    typeName = typeName,
    providerTypeName = typeName.wrapInProvider(),
    lazyTypeName = typeName.wrapInLazy(),
    isWrappedInProvider = isWrappedInProvider,
    isWrappedInLazy = isWrappedInLazy,
    isLazyWrappedInProvider = isLazyWrappedInProvider,
    isAssisted = assistedAnnotation != null,
    assistedIdentifier = assistedIdentifier
  )
}

internal fun FunSpec.Builder.addMemberInjection(
  memberInjectParameters: List<MemberInjectParameter>,
  instanceName: String
): FunSpec.Builder = apply {

  memberInjectParameters.forEach { parameter ->

    val functionName = "inject${parameter.accessName.capitalize()}"

    val param = when {
      parameter.isLazyWrappedInProvider ->
        "${ProviderOfLazy::class.qualifiedName}.create(${parameter.name})"
      parameter.isWrappedInProvider -> parameter.name
      parameter.isWrappedInLazy ->
        "$daggerDoubleCheckFqNameString.lazy(${parameter.name})"
      else -> parameter.name + ".get()"
    }

    addStatement("%T.$functionName($instanceName, $param)", parameter.memberInjectorClassName)
  }
}

/**
 * Returns all member-injected parameters for the receiver class *and any superclasses*.
 *
 * We use Psi whenever possible, to support generated code.
 *
 * Order is important. Dagger expects the properties of the most-upstream class to be listed first
 * in a factory's constructor.
 *
 * Given the hierarchy:
 * Impl -> Middle -> Base
 * The order of dependencies in `Impl_Factory`'s constructor should be:
 * Base -> Middle -> Impl
 */
internal fun ClassReference.memberInjectParameters(): List<MemberInjectParameter> {
  return allSuperTypeClassReferences(includeSelf = true)
    .filterNot { it.isInterface() }
    .toList()
    .foldRight(listOf()) { classReference, acc ->
      acc + classReference.declaredMemberInjectParameters(acc)
    }
}

/**
 * @param superParameters injected parameters from any super-classes, regardless of whether they're
 * overridden by the receiver class
 * @return the member-injected parameters for this class only, not including any super-classes
 */
private fun ClassReference.declaredMemberInjectParameters(
  superParameters: List<Parameter>
): List<MemberInjectParameter> {
  return properties
    .filter { it.isAnnotatedWith(injectFqName) }
    .filter { it.visibility() != PRIVATE }
    .fold(listOf()) { acc, property ->
      val uniqueName = property.name.uniqueParameterName(superParameters, acc)
      acc + property.toMemberInjectParameter(uniqueName = uniqueName)
    }
}

/**
 * Converts the parameter list to comma separated argument list that can be used to call other
 * functions, e.g.
 * ```
 * [param0: String, param1: Int] -> "param0, param1"
 * ```
 * [asProvider] allows you to decide if each parameter is wrapped in a `Provider` interface. If
 * true, then the `get()` function will be called for the provider parameter. If false, then
 * then always only the parameter name will used in the argument list:
 * ```
 * "param0.get()" vs "param0"
 * ```
 * Set [includeModule] to true if a Dagger module instance is part of the argument list.
 */
internal fun List<Parameter>.asArgumentList(
  asProvider: Boolean,
  includeModule: Boolean
): String {
  return this
    .let { list ->
      if (asProvider) {
        list.map { parameter ->
          when {
            parameter.isLazyWrappedInProvider ->
              "${ProviderOfLazy::class.qualifiedName}.create(${parameter.name})"
            parameter.isWrappedInProvider -> parameter.name
            // Normally Dagger changes Lazy<Type> parameters to a Provider<Type> (usually the
            // container is a joined type), therefore we use `.lazy(..)` to convert the Provider
            // to a Lazy. Assisted parameters behave differently and the Lazy type is not changed
            // to a Provider and we can simply use the parameter name in the argument list.
            parameter.isWrappedInLazy && parameter.isAssisted -> parameter.name
            parameter.isWrappedInLazy -> "$daggerDoubleCheckFqNameString.lazy(${parameter.name})"
            parameter.isAssisted -> parameter.name
            else -> "${parameter.name}.get()"
          }
        }
      } else list.map { it.name }
    }
    .let {
      if (includeModule) {
        val result = it.toMutableList()
        result.add(0, "module")
        result.toList()
      } else {
        it
      }
    }
    .joinToString()
}

private fun PropertyReference.toMemberInjectParameter(
  uniqueName: String
): MemberInjectParameter {
  val originalName = name
  val type = type()

  val isWrappedInProvider = type.asClassReferenceOrNull()?.fqName == providerFqName
  val isWrappedInLazy = type.asClassReferenceOrNull()?.fqName == daggerLazyFqName
  val isLazyWrappedInProvider = isWrappedInProvider &&
    type.unwrappedFirstType.asClassReferenceOrNull()?.fqName == daggerLazyFqName

  val typeName = when {
    isLazyWrappedInProvider -> type.unwrappedFirstType.unwrappedFirstType
    isWrappedInProvider || isWrappedInLazy -> type.unwrappedFirstType
    else -> type
  }.asTypeName().withJvmSuppressWildcardsIfNeeded(this, type)

  val assistedAnnotation = annotations.singleOrNull { it.fqName == assistedFqName }

  val assistedIdentifier = assistedAnnotation
    ?.argumentAt("value", 0)
    ?.value()
    ?: ""

  val memberInjectorClassName = declaringClass
    .generateClassName(separator = "_", suffix = "_MembersInjector")
    .relativeClassName
    .asString()

  val memberInjectorClass = ClassName(
    declaringClass.packageFqName.asString(), memberInjectorClassName
  )

  val isSetterInjected = this.setterAnnotations.any { it.fqName == injectFqName }

  // setter delegates require a "set" prefix for their inject function
  val accessName = if (isSetterInjected) {
    "set${originalName.capitalize()}"
  } else {
    originalName
  }

  val qualifierAnnotations = annotations
    .filter { it.isQualifier() }
    .map { it.toAnnotationSpec() }

  return MemberInjectParameter(
    name = uniqueName,
    originalName = originalName,
    typeName = typeName,
    providerTypeName = typeName.wrapInProvider(),
    lazyTypeName = typeName.wrapInLazy(),
    isWrappedInProvider = isWrappedInProvider,
    isWrappedInLazy = isWrappedInLazy,
    isLazyWrappedInProvider = isLazyWrappedInProvider,
    isAssisted = assistedAnnotation != null,
    assistedIdentifier = assistedIdentifier,
    memberInjectorClassName = memberInjectorClass,
    isSetterInjected = isSetterInjected,
    accessName = accessName,
    qualifierAnnotationSpecs = qualifierAnnotations,
    injectedFieldSignature = fqName
  )
}

internal fun ClassName.optionallyParameterizedBy(
  typeParameters: List<TypeParameterReference>
): TypeName {
  return if (typeParameters.isEmpty()) {
    this
  } else {
    parameterizedBy(typeParameters.map { it.typeVariableName })
  }
}
